# import dill
from PIL import Image
import os
import subprocess
import time
import json
import bson
import numpy as np
from .inverse_design import *
from .sparams import *
from .utils import *
from .layers import *
from .constants import *


from subprocess import Popen, PIPE


def solve(prob, dev=False, run=True):
    if "dev" in prob:
        dev = prob["dev"]
    # c0 = prob["component"]
    # del prob["component"]
    bson_data = bson.dumps(prob)
    # prob["component"] = c0

    path = prob["path"]
    if not os.path.exists(path):
        os.makedirs(path)
    print(f"""
          using simulation folder {path}
          started julia process
          compiling julia code...
          """)
    prob_path = os.path.join(path, "prob.bson")
    with open(prob_path, "wb") as f:
        # Write the BSON data to the file
        f.write(bson_data)

    if not run:
        return
    start_time = time.time()

    def run(cmd):

        proc = Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # proc.wait()
        with proc:
            for line in proc.stdout:

                print(str(line.decode().strip()), flush=True)
            err_message = proc.stderr.read().decode()
            print(err_message)
    cmd_dev = [f"julia", os.path.join(os.path.dirname(
        os.path.abspath(__file__)), "run.jl"), path,]
    cmd = ["lumi", path]
    if dev:
        print(" ".join(cmd_dev))
        run(cmd_dev)
    else:
        try:
            run(cmd)
        except Exception as e:
            print(e)
            run(cmd_dev)

    # with Popen(cmd,  stdout=PIPE, stderr=PIPE) as p:
    #     if p.stderr is not None:
    #         for line in p.stderr:
    #             print(line, flush=True)
    # exit_code = p.poll()
    # subprocess.run()
    # print(f"julia simulation took {time.time()-start_time} seconds")
    print(f"images and results saved in {path}")
    sol = load_solution(path=path)
    # if prob["study"] == "inverse_design":
    # c = apply_design(c0,  sol)
    # sol["optimized_component"] = c
    #     sol["before"]["component"] = c0
    #     sol["after"]["component"] = c
    return sol


def get_path(path=None, study=""):
    if path is None:
        l = sorted(os.listdir(RUNS_PATH), reverse=True)
        if study:
            for p in l:
                try:
                    s = json.loads(open(os.path.join(RUNS_PATH, p, "sol.json")).read())[
                        "study"]
                    if s == study:
                        path = p
                        break
                except:
                    pass
        else:
            path = l[0]
        path = os.path.join(RUNS_PATH, path)
        return path
    if os.path.isdir(path):
        return path

    if os.path.isdir(os.path.join(RUNS_PATH, path)):
        return path


def load_component(path=None, study=None):
    return dill.load(os.path.join(get_path(path), "comp.bson"))


def finetune(iters, path=None,):
    if path is None:
        path = get_path(study="inverse_design")

    prob = bson.loads(open(os.path.join(path, "prob.bson"), "rb").read())
    prob["iters"] = iters
    # with open(path, "wb") as f:
    #     f.write(bson.dumps(prob))
    solve(prob)


def load_sparams(sparams):
    if "re" in list(sparams.values())[0]:
        return {k: v["re"]+1j*v["im"] for k, v in sparams.items()}
    return {wl: {k: (v["re"]+1j*v["im"])
                 for k, v in d.items()} for wl, d in sparams.items()}


def load_solution(path=None, study="",):
    if path is None:
        path = get_path(study=study)
    print(f"loading solution from {path}")
    prob = bson.loads(open(os.path.join(path, "prob.bson"), "rb").read())
    p = os.path.join(path, "sol.json")
    # sol = bson.loads(p, "rb").read())["sol"]
    sol = json.loads(open(p).read())
    sol["sparams"] = load_sparams(sol["sparams"])
    if prob["study"] == "sparams":
        pass
    elif prob["study"] == "inverse_design":
        # for k in ["before", "after"]:
        #     sol[k]["sparams"] = load_sparams(sol[k]["sparams"])
        # for k in sol["after"]:
        #     sol[k] = sol["after"][k]

        l = [np.array(d) for d in sol["optimized_designs"]]
        sol["optimized_designs"] = l
        for i, a in enumerate(l):
            Image.fromarray(np.uint8(a * 255),
                            'L').save(os.path.join(path, f"design{i+1}.png"))
            pic2gds(os.path.join(path, f"design{i+1}.png"), sol["dx"])
        pass
    return sol


def write_sparams(*args, run=True, **kwargs):
    return solve(sparams_problem(*args, **kwargs), run=run)





class Si():
    epsilon = 3.4757**2


class SiO2():
    epsilon = 1.444**2


class SiN():
    epsilon = 2.0**2


class Ge():
    epsilon = 4.2**2


MATERIALS = {
    "si": Si,
    "sio2": SiO2,
    "Si": Si,
    "SiO2": SiO2,
    "SiN": SiN,
    "sin": SiN,
    "Ge": Ge,
    "ge": Ge,

}
# Ge["epsilon"]




from .constants import *
from .utils import *
import bson
import gdsfactory as gf
from gdsfactory.cross_section import Section
from networkx import center
from numpy import cumsum
from gdsfactory.generic_tech import LAYER_STACK, LAYER

# import .utils as utils
# from layers import *


def mimo(west=0, east=0, south=0, north=0,
         l=2.0, w=2.0, wwg=.5, lwg=None,
         wwg_west=None, wwg_east=None, wwg_south=None, wwg_north=None,
         wwg_layer=LAYER.WG,  # bbox_layer=LAYER.WAFER,
         design_layer=DESIGN_LAYER,
         **kwargs):
    design = gf.Component()

    c = gf.Component()
    if lwg is None:
        lwg = 4*wwg
    p = [(0, 0), (l, 0), (l, w), (0, w)]
    design.add_polygon(p,                       layer=design_layer)
    c.add_polygon(p,                       layer=wwg_layer)

    ld = [west,  east, south, north]
    for i, v, d in zip(range(4), ld, [w, w, l, l]):
        if type(v) is int:
            ld[i] = [(.5+j)*d/v for j in range(v)]
    lwwg = [wwg_west, wwg_east, wwg_south, wwg_north]
    for i, v in enumerate(lwwg):
        if v is None:
            v = wwg
        if type(v) is float or type(v) is int:
            lwwg[i] = [v]*len(ld[i])

    n = 0
    for (i, x, y, d, wwg, a) in zip(
        range(4),
        [0,  l, 0, 0],
        [0, 0, 0, w],
        ld,
        lwwg,
        [180, 0, -90, 90]
    ):
        for wwg, v in zip(wwg, d):
            center = (x, y+v) if i in [0, 1] else (x+v, y)
            name = "o"+str(n+1)
            design.add_port(name=name, center=center, width=wwg,
                            orientation=a, layer=wwg_layer)
            wg = c << gf.components.straight(
                length=lwg, width=wwg, layer=wwg_layer)
            wg.connect("o2", design.ports[name])
            c.add_port(name=name, port=wg.ports["o1"])
            n += 1

    design = c << design
    return c


# def ubend_template(r, l, wwg, lwg=0,
    #                wavelength=1.55, dx=0.05,
    #                #    ϵcore=ϵcore, ϵclad=ϵclad,
    #                dir="", init=None, lmin=0.1, LAYER=LAYER, **kwargs):
    # name = "ubend"
    # design = gf.Component("design")
    # device = gf.Component("ubend")
    # λ = wavelength

    # if lwg == 0:
    #     lwg = 4*wwg

    # # Extrude the Path and the CrossSection
    # wg1 = device << gf.path.extrude(gf.path.straight(
    #     length=lwg), layer=LAYER.WG, width=wwg)
    # wg2 = device << gf.path.extrude(gf.path.straight(
    #     length=lwg), layer=LAYER.WG, width=wwg)

    # ld = l
    # wd = 2*r+wwg

    # design.add_polygon([(0, -wd/2), (ld, -wd/2), (ld, wd/2),
    #                     (0, wd/2)], layer=DESIGN)
    # if init is not None:
    #     xmin = design.xmin
    #     ymin = design.ymin
    #     init = design << init
    #     init.xmin = xmin
    #     init.ymin = ymin
    # else:
    #     init = gf.Component()
    #     init.add_polygon([(0, -wd/2), (ld, -wd/2), (ld, wd/2),
    #                       (0, wd/2)], layer=GUESS)
    #     init = design << init

    # design.add_port(name="o1", center=(0, r), width=wwg,
    #                 orientation=180, layer=LAYER.PORT)
    # design.add_port(name="o2", center=(0, -r), width=wwg,
    #                 orientation=180, layer=LAYER.PORT)
    # # design = device << design

    # wg1.connect("o2", design.ports["o1"], allow_layer_mismatch=True)
    # wg2.connect("o2", design.ports["o2"], allow_layer_mismatch=True)
    # # stub1.connect("o2", wg1.ports["o1"])
    # # stub2.connect("o2", wg2.ports["o1"])

    # device.add_port("o1", port=wg1.ports["o1"])
    # device.add_port("o2", port=wg2.ports["o1"])

    # # design = device << design
    # device.add_ref(design, name="design")
    # # device << gf.components.re
    # # device << test
    # device.show()
    # # init.show()
    # sources = {
    #     "o1": {
    #         "port": "o1",
    #         "wavelengths": [λ],
    #         "mode_numbers": [0],
    #         "powers": [1],
    #         "width": wwg,
    #         "endpoints": device.ports["o1"].endpoints.tolist(),
    #         "center": (device.ports["o1"].center),
    #         "orientation": device.ports["o1"].orientation,
    #     }}
    # for k in sources:
    #     a = sources[k]["orientation"]/180*pi
    #     normal = [cos(a), sin(a)]
    #     sources[k]["normal"] = normal
    #     sources[k]["center"] -= .001*np.array(normal)
    #     sources[k]["center"] = sources[k]["center"].flatten().tolist()

    # monitors = {
    #     "o2":        {
    #         "powers": [1],
    #         "wavelengths": [λ],
    #         "mode_numbers": [0],
    #         "port": "o2"},
    #     # {"power": 1,"wavelength": λ,"port": "o1"},
    # }
    # # device.metadata["sources"] = sources
    # # device.metadata["monitors"] = monitors
    # return device, sources, monitors, design



# import gplugins.luminescent as gl
from . import gcells
from .utils import *
from .constants import *
from .materials import MATERIALS
from .inverse_design import *
from .sparams import *
from .sol import *
from .setup import *
from gdsfactory.generic_tech import (
    LAYER_STACK, LAYER

)

__all__ = [
    "LAYER_STACK",
    "LAYER",
    "LAYER_VIEWS",
    "MATERIALS",
    "write_sparams",
    "inverse_design_problem",
    "sparams_problem",
    "solve",
    "finetune",
    "apply_design",
    "add_bbox",
    "XYMARGIN",
    "XMARGIN",
    "YMARGIN",
    "ZMARGIN",
    "load_solution",
    "load_component",
    "show_solution",
    "gcells",
]
